﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows;

using Size = System.Windows.Size;

namespace TestLibrary
{
    /// <summary>
    /// 스트림 재생기 컨트롤
    /// </summary>
    public partial class StreamPlayerControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Delegate
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 스트림 실패시 이벤트 핸들러 - StreamFailedEventHandler(sender, e)

        /// <summary>
        /// 스트림 실패시 이벤트 핸들러
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        public delegate void StreamFailedEventHandler(object sender, StreamFailedEventArgs e);

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 콜백 대리자 - CallbackDelegate()

        /// <summary>
        /// 콜백 대리자
        /// </summary>
        [UnmanagedFunctionPointer(CallingConvention.StdCall, SetLastError = true)]
        private delegate void CallbackDelegate();

        #endregion
        #region 에러 콜백 대리자 - ErrorCallbackDelegate(error)

        /// <summary>
        /// 에러 콜백 대리자
        /// </summary>
        /// <param name="error">에러</param>
        [UnmanagedFunctionPointer(CallingConvention.StdCall, SetLastError = true)]
        private delegate void ErrorCallbackDelegate(string error);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 스트림 시작시 이벤트 - StreamStarted

        /// <summary>
        /// 스트림 시작시 이벤트
        /// </summary>
        public event RoutedEventHandler StreamStarted
        {
            add
            {
                AddHandler(StreamStartedEvent, value);
            }
            remove
            {
                RemoveHandler(StreamStartedEvent, value);
            }
        }

        #endregion
        #region 스트림 중단시 이벤트 - StreamStopped

        /// <summary>
        /// 스트림 중단시 이벤트
        /// </summary>
        public event RoutedEventHandler StreamStopped
        {
            add
            {
                AddHandler(StreamStoppedEvent, value);
            }
            remove
            {
                RemoveHandler(StreamStoppedEvent, value);
            }
        }

        #endregion
        #region 스트림 실패시 이벤트 - StreamFailed

        /// <summary>
        /// 스트림 실패시 이벤트
        /// </summary>
        public event StreamFailedEventHandler StreamFailed
        {
            add
            {
                AddHandler(StreamFailedEvent, value);
            }
            remove
            {
                RemoveHandler(StreamFailedEvent, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 재생 여부 속성 키
        /// </summary>
        private static readonly DependencyPropertyKey IsPlayingPropertyKey = DependencyProperty.RegisterReadOnly
        (
            "IsPlaying",
            typeof(bool),
            typeof(StreamPlayerControl),
            new FrameworkPropertyMetadata(false)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 스트림 시작시 이벤트
        /// </summary>
        public static readonly RoutedEvent StreamStartedEvent = EventManager.RegisterRoutedEvent
        (
            "StreamStarted",
            RoutingStrategy.Bubble,
            typeof(RoutedEventHandler),
            typeof(StreamPlayerControl)
        );

        /// <summary>
        /// 스트림 중단시 이벤트
        /// </summary>
        public static readonly RoutedEvent StreamStoppedEvent = EventManager.RegisterRoutedEvent
        (
            "StreamStopped",
            RoutingStrategy.Bubble,
            typeof(RoutedEventHandler),
            typeof(StreamPlayerControl)
        );

        /// <summary>
        /// 스트림 실패시 이벤트
        /// </summary>
        public static readonly RoutedEvent StreamFailedEvent = EventManager.RegisterRoutedEvent
        (
            "StreamFailed",
            RoutingStrategy.Bubble,
            typeof(StreamFailedEventHandler),
            typeof(StreamPlayerControl)
        );

        /// <summary>
        /// 재생 여부 속성
        /// </summary>
        public static readonly DependencyProperty IsPlayingProperty = IsPlayingPropertyKey.DependencyProperty;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 스트림 재생기 헬퍼
        /// </summary>
        private StreamPlayerHelper helper;

        /// <summary>
        /// 스트림 시작시 콜백 대리자
        /// </summary>
        private Delegate streamStartedCallbackDelegate;

        /// <summary>
        /// 스트림 중단시 콜백 대리자
        /// </summary>
        private Delegate streamStoppedCallbackDelegate;

        /// <summary>
        /// 스트림 실패시 콜백 대리자
        /// </summary>
        private Delegate streamFailedCallbackDelegate;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 스트램 재생시 헬퍼 - Helper

        /// <summary>
        /// 스트램 재생시 헬퍼
        /// </summary>
        private StreamPlayerHelper Helper
        {
            get
            {
                return this.helper;
            }
        }

        #endregion
        #region 재생 여부 - IsPlaying

        /// <summary>
        /// 재생 여부
        /// </summary>
        [Browsable(false)]
        public bool IsPlaying
        {
            get
            {
                return (bool)GetValue(IsPlayingProperty);
            }
            private set
            {
                SetValue(IsPlayingPropertyKey, value);
            }
        }

        #endregion
        #region 비디오 크기 - VideoSize

        /// <summary>
        /// 비디오 크기
        /// </summary>
        [Browsable(false)]
        public Size VideoSize
        {
            get
            {
                return IsPlaying ? this.helper.GetFrameSize() : new Size(0, 0);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - StreamPlayerControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public StreamPlayerControl()
        {
            InitializeComponent();

            Loaded += UserControl_Loaded;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 재생 시작하기 - StartPlay(uri, connectionTimeout, transport, flag)

        /// <summary>
        /// 재생 시작하기
        /// </summary>
        /// <param name="uri">URI</param>
        /// <param name="connectionTimeout">연결 타임아웃</param>
        /// <param name="transport">RTSP 전송</param>
        /// <param name="flag">RTSP 플래그</param>
        public void StartPlay(Uri uri, TimeSpan connectionTimeout, RTSPTransport transport, RTSPFlag flag)
        {
            if(IsPlaying)
            {
                Stop();
            }

            this.helper.StartPlay(uri.IsFile ? uri.LocalPath : uri.ToString(), connectionTimeout, transport, flag);
        }

        #endregion
        #region 재생 시작하기 - StartPlay(uri)

        /// <summary>
        /// 재생 시작하기
        /// </summary>
        /// <param name="uri">URI</param>
        public void StartPlay(Uri uri)
        {
            StartPlay(uri, TimeSpan.FromSeconds(5.0), RTSPTransport.Undefined, RTSPFlag.None);
        }

        #endregion
        #region 현재 프레임 구하기 - GetCurrentFrame()

        /// <summary>
        /// 현재 프레임 구하기
        /// </summary>
        /// <returns>비트맵</returns>
        public Bitmap GetCurrentFrame()
        {
            if(!IsPlaying)
            {
                throw new InvalidOperationException();
            }

            return this.helper.GetCurrentFrame();
        }

        #endregion
        #region 중단하기 - Stop()

        /// <summary>
        /// 중단하기
        /// </summary>
        public void Stop()
        {
            if(!IsPlaying)
            {
                throw new InvalidOperationException();
            }

            this.helper.Stop();

            IsPlaying = false;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 컨텐트 변경시 처리하기 - OnContentChanged(oldContent, newContent)

        /// <summary>
        /// 컨텐트 변경시 처리하기
        /// </summary>
        /// <param name="oldContent">이전 컨텐트</param>
        /// <param name="newContent">신규 컨텐트</param>
        protected override void OnContentChanged(object oldContent, object newContent)
        {
            if(oldContent != null)
            {
                throw new InvalidOperationException();
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 사용자 컨트롤 로드시 처리하기 - UserControl_Loaded(sender, e)

        /// <summary>
        /// 사용자 컨트롤 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void UserControl_Loaded(object sender, RoutedEventArgs e)
        {
            this.helper = GetStreamPlayerHelper();

            Window window = Window.GetWindow(this);

            if(window != null)
            {
                window.Closed += Window_Closed;

                Loaded -= UserControl_Loaded;
            }
        }

        #endregion
        #region 윈도우 닫는 경우 처리하기 - Window_Closed(sender, e)

        /// <summary>
        /// 윈도우 닫는 경우 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Window_Closed(object sender, EventArgs e)
        {
            if(this.helper != null)
            {
                if(IsPlaying)
                {
                    Stop();
                }

                this.helper.Uninitialize();

                this.helper.Dispose();
            }

            Window window = sender as Window;

            if(window != null)
            {
                window.Closed -= Window_Closed;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 스트림 재생기 헬퍼 구하기 - GetStreamPlayerHelper()

        /// <summary>
        /// 스트림 재생기 헬퍼 구하기
        /// </summary>
        /// <returns>스트림 재생기 헬퍼</returns>
        private StreamPlayerHelper GetStreamPlayerHelper()
        {
            StreamPlayerHelper helper = new StreamPlayerHelper();

            this.streamStartedCallbackDelegate = new CallbackDelegate(FireStreamStartedEvent);
            this.streamStoppedCallbackDelegate = new CallbackDelegate(FireStreamStoppedEvent);
            this.streamFailedCallbackDelegate  = new ErrorCallbackDelegate(FireStreamFailedEvent);
            
            StreamPlayerParameter parameter = new StreamPlayerParameter
            {
                WindowHandle                = this.videoWindow.Handle,
                StreamStartedCallbackHandle = Marshal.GetFunctionPointerForDelegate(streamStartedCallbackDelegate),
                StreamStoppedCallbackHandle = Marshal.GetFunctionPointerForDelegate(streamStoppedCallbackDelegate),
                StreamFailedCallbackHandle  = Marshal.GetFunctionPointerForDelegate(streamFailedCallbackDelegate)
            };

            helper.Initialize(parameter);

            return helper;
        }

        #endregion

        #region 스트림 시작시 이벤트 발생시키기 - FireStreamStartedEvent()

        /// <summary>
        /// 스트림 시작시 이벤트 발생시키기
        /// </summary>
        private void FireStreamStartedEvent()
        {
            IsPlaying = true;

            RaiseEvent(new RoutedEventArgs(StreamStartedEvent));
        }

        #endregion
        #region 스트림 중단시 이벤트 발생시키기 - FireStreamStoppedEvent()

        /// <summary>
        /// 스트림 중단시 이벤트 발생시키기
        /// </summary>
        private void FireStreamStoppedEvent()
        {
            IsPlaying = false;

            RaiseEvent(new RoutedEventArgs(StreamStoppedEvent));
        }

        #endregion
        #region 스트림 실패시 이벤트 발생시키기 - FireStreamFailedEvent(error)

        /// <summary>
        /// 스트림 실패시 이벤트 발생시키기
        /// </summary>
        /// <param name="error">에러</param>
        private void FireStreamFailedEvent(string error)
        {
            IsPlaying = false;

            RaiseEvent(new StreamFailedEventArgs(StreamFailedEvent, error));
        }

        #endregion
    }
}