﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;

using Size = System.Windows.Size;

using TestLibrary.Properties;

namespace TestLibrary
{
    /// <summary>
    /// 스트림 재생기 헬퍼
    /// </summary>
    public sealed class StreamPlayerHelper : IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Import
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 라이브러리 로드하기 - LoadLibrary(filePath)

        /// <summary>
        /// 라이브러리 로드하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <returns>처리 결과</returns>
        [DllImport("kernel32", SetLastError = true, CharSet = CharSet.Unicode)]
        private static extern IntPtr LoadLibrary(string filePath);

        #endregion
        #region 라이브러리 해제하기 - FreeLibrary(moduleHandle)

        /// <summary>
        /// 라이브러리 해제하기
        /// </summary>
        /// <param name="moduleHandle">모듈 핸들</param>
        /// <returns>처리 결과</returns>
        [DllImport("kernel32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool FreeLibrary(IntPtr moduleHandle);

        #endregion
        #region 프로시저 주소 구하기 - GetProcAddress(moduleHandle, procedureName)

        /// <summary>
        /// 프로시저 주소 구하기
        /// </summary>
        /// <param name="moduleHandle">모듈 핸들</param>
        /// <param name="procedureName">프로시저명</param>
        /// <returns>처리 결과</returns>
        [DllImport("kernel32", CharSet = CharSet.Ansi, ExactSpelling = true, SetLastError = true)]
        private static extern IntPtr GetProcAddress(IntPtr moduleHandle, string procedureName);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Delegate
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 초기화하기 대리자 - InitializeDelegate(parameter)

        /// <summary>
        /// 초기화하기 대리자
        /// </summary>
        /// <param name="parameter">매개 변수</param>
        /// <returns>처리 결과</returns>
        private delegate int InitializeDelegate(StreamPlayerParameter parameter);

        #endregion
        #region 재생 시작하기 대리자 - StartPlayDelegate(url, connectionTimeout, transport, flag)

        /// <summary>
        /// 재생 시작하기 대리자
        /// </summary>
        /// <param name="url">URL</param>
        /// <param name="connectionTimeout">연결 타임아웃</param>
        /// <param name="transport">전송</param>
        /// <param name="flag">플래그</param>
        /// <returns>처리 결과</returns>
        private delegate int StartPlayDelegate([MarshalAs(UnmanagedType.LPStr)]string url, int connectionTimeout, int transport, int flag);

        #endregion
        #region 프레임 크기 구하기 대리자 - GetFrameSizeDelegate(width, height)

        /// <summary>
        /// 프레임 크기 구하기 대리자
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>처리 결과</returns>
        private delegate int GetFrameSizeDelegate(out int width, out int height);

        #endregion
        #region 현재 프레임 구하기 대리자 - GetCurrentFrameDelegate(dibHandle)

        /// <summary>
        /// 현재 프레임 구하기 대리자
        /// </summary>
        /// <param name="dibHandle">DIB 핸들</param>
        /// <returns>처리 결과</returns>
        private delegate int GetCurrentFrameDelegate([Out] out IntPtr dibHandle);

        #endregion
        #region 중단하기 대리자 - StopDelegate()

        /// <summary>
        /// 중단하기 대리자
        /// </summary>
        /// <returns>처리 결과</returns>
        private delegate int StopDelegate();

        #endregion
        #region 초기화 취소하기 대리자 - UninitializeDelegate()

        /// <summary>
        /// 초기화 취소하기 대리자
        /// </summary>
        /// <returns>처리 결과</returns>
        private delegate int UninitializeDelegate();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 모듈 파일 경로
        /// </summary>
        private string moduleFilePath = string.Empty;

        /// <summary>
        /// 모듈 핸들
        /// </summary>
        private IntPtr moduleHandle = IntPtr.Zero;

        /// <summary>
        /// 초기화하기 대리자
        /// </summary>
        private InitializeDelegate initializeDelegate;

        /// <summary>
        /// 재생 시작하기 대리자
        /// </summary>
        private StartPlayDelegate startPlayDelegate;

        /// <summary>
        /// 프레임 크기 구하기 대리자
        /// </summary>
        private GetFrameSizeDelegate getFrameSizeDelegate;

        /// <summary>
        /// 현재 프레임 구하기 대리자
        /// </summary>
        private GetCurrentFrameDelegate getCurrentFrameDelegate;

        /// <summary>
        /// 중단하기 대리자
        /// </summary>
        private StopDelegate stopDelegate;

        /// <summary>
        /// 초기화 취소하기 대리자
        /// </summary>
        private UninitializeDelegate uninitializeDelegate;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region X86 플랫폼 여부 - IsX86Platform

        /// <summary>
        /// X86 플랫폼 여부
        /// </summary>
        private bool IsX86Platform
        {
            get
            {
                return IntPtr.Size == 4;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - StreamPlayerHelper()

        /// <summary>
        /// 생성자
        /// </summary>
        public StreamPlayerHelper()
        {
            LoadModule();

            SetDelegate(this.moduleHandle);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 초기화하기 - Initialize(parameter)

        /// <summary>
        /// 초기화하기
        /// </summary>
        /// <param name="parameter">스트림 재생기 매개 변수</param>
        public void Initialize(StreamPlayerParameter parameter)
        {
            if(this.initializeDelegate(parameter) != 0)
            {
                throw new StreamPlayerException("Failed to initialize the player.");
            }
        }

        #endregion
        #region 재생 시작하기 - StartPlay(url, connectionTimeout, transport, flag)

        /// <summary>
        /// 재생 시작하기
        /// </summary>
        /// <param name="url">URL</param>
        /// <param name="connectionTimeout">연결 타임아웃</param>
        /// <param name="transport">RTSP 전송</param>
        /// <param name="flag">RTSP 플래그</param>
        public void StartPlay(string url, TimeSpan connectionTimeout, RTSPTransport transport, RTSPFlag flag)
        {
            if(this.startPlayDelegate(url, Convert.ToInt32(connectionTimeout.TotalMilliseconds), Convert.ToInt32(transport), Convert.ToInt32(flag)) != 0)
            {
                throw new StreamPlayerException("Failed to play the stream.");
            }
        }

        #endregion
        #region 프레임 크기 구하기 - GetFrameSize()

        /// <summary>
        /// 프레임 크기 구하기
        /// </summary>
        /// <returns>프레임 크기</returns>
        public Size GetFrameSize()
        {
            int width;
            int height;

            if(this.getFrameSizeDelegate(out width, out height) != 0)
            {
                throw new StreamPlayerException("Failed to get the frame size.");
            }

            return new Size(width, height);
        }

        #endregion
        #region 현재 프레임 구하기 - GetCurrentFrame()

        /// <summary>
        /// 현재 프레임 구하기
        /// </summary>
        /// <returns>비트맵</returns>
        public Bitmap GetCurrentFrame()
        {
            IntPtr dibHandle;

            if(this.getCurrentFrameDelegate(out dibHandle) != 0)
            {
                throw new StreamPlayerException("Failed to get the current image.");
            }

            try
            {
                BITMAPINFOHEADER header = (BITMAPINFOHEADER)Marshal.PtrToStructure(dibHandle, typeof(BITMAPINFOHEADER));

                int stride = header.Width * (header.BitCount / 8);

                int padding = stride % 4 > 0 ? 4 - stride % 4 : 0;

                stride += padding;

                Bitmap bitmap = new Bitmap
                (
                    header.Width,
                    header.Height,
                    stride,
                    PixelFormat.Format24bppRgb,
                    (IntPtr)(dibHandle.ToInt64() + Marshal.SizeOf(header))
                );

                bitmap.RotateFlip(RotateFlipType.RotateNoneFlipY);

                return bitmap;
            }
            finally
            {
                if(dibHandle != IntPtr.Zero)
                {
                    Marshal.FreeCoTaskMem(dibHandle);
                }
            }
        }

        #endregion
        #region 중단하기 - Stop()

        /// <summary>
        /// 중단하기
        /// </summary>
        public void Stop()
        {
            if(this.stopDelegate() != 0)
            {
                throw new StreamPlayerException("Failed to stop the stream.");
            }
        }

        #endregion
        #region 초기화 취소하기 - Uninitialize()

        /// <summary>
        /// 초기화 취소하기
        /// </summary>
        public void Uninitialize()
        {
            if(this.uninitializeDelegate() != 0)
            {
                throw new StreamPlayerException("Failed to uninitialize the player.");
            }
        }

        #endregion

        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            if(this.moduleHandle != IntPtr.Zero)
            {
                FreeLibrary(this.moduleHandle);

                this.moduleHandle = IntPtr.Zero;
            }

            if(File.Exists(this.moduleFilePath))
            {
                 File.Delete(this.moduleFilePath);
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 모듈 로드하기 - LoadModule()

        /// <summary>
        /// 모듈 로드하기
        /// </summary>
        private void LoadModule()
        {
            this.moduleFilePath = Path.GetTempFileName();

            using(FileStream stream = new FileStream(this.moduleFilePath, FileMode.Create, FileAccess.Write))
            {
                using(BinaryWriter writer = new BinaryWriter(stream))
                {
                    writer.Write(IsX86Platform ? Resources.StreamPlayer : Resources.StreamPlayer64);
                }
            }

            this.moduleHandle = LoadLibrary(this.moduleFilePath);

            if(this.moduleHandle == IntPtr.Zero)
            {
                throw new Win32Exception(Marshal.GetLastWin32Error());
            }
        }

        #endregion
        #region 대리자 설정하기 - SetDelegate(moduleHandle)

        /// <summary>
        /// 대리자 설정하기
        /// </summary>
        /// <param name="moduleHandle">모듈 핸들</param>
        private void SetDelegate(IntPtr moduleHandle)
        {
            IntPtr procedureHandle = GetProcAddress(moduleHandle, "Initialize");

            this.initializeDelegate = (InitializeDelegate)Marshal.GetDelegateForFunctionPointer(procedureHandle, typeof(InitializeDelegate));

            procedureHandle = GetProcAddress(moduleHandle, "StartPlay");

            this.startPlayDelegate = (StartPlayDelegate)Marshal.GetDelegateForFunctionPointer(procedureHandle, typeof(StartPlayDelegate));

            procedureHandle = GetProcAddress(moduleHandle, "GetFrameSize");

            this.getFrameSizeDelegate = (GetFrameSizeDelegate)Marshal.GetDelegateForFunctionPointer(procedureHandle, typeof(GetFrameSizeDelegate));

            procedureHandle = GetProcAddress(moduleHandle, "GetCurrentFrame");

            this.getCurrentFrameDelegate = (GetCurrentFrameDelegate)Marshal.GetDelegateForFunctionPointer(procedureHandle, typeof(GetCurrentFrameDelegate));

            procedureHandle = GetProcAddress(moduleHandle, "Stop");

            this.stopDelegate = (StopDelegate)Marshal.GetDelegateForFunctionPointer(procedureHandle, typeof(StopDelegate));

            procedureHandle = GetProcAddress(moduleHandle, "Uninitialize");

            this.uninitializeDelegate = (UninitializeDelegate)Marshal.GetDelegateForFunctionPointer(procedureHandle, typeof(UninitializeDelegate));
        }

        #endregion
    }
}